#include "system.h"
#include "util.h"
#include "type.h"
#include "network.h"
#include "connect.h"
#include "example.h"
#include "train.h"
#include "act.h"
#include "command.h"
#include "control.h"
#include "object.h"

/* Add your initials or something here to customize the version number */
char *VersionExt = "CTK";

/******************************** Extension Records **************************/

ObjInfo NetExtInfo;
ObjInfo GroupExtInfo;
ObjInfo UnitExtInfo;
ObjInfo BlockExtInfo;
ObjInfo ExSetExtInfo;
ObjInfo ExampleExtInfo;
ObjInfo EventExtInfo;

/* This is what the object viewer displays as the name of the extension. */
static void extensionName(void *ext, char *dest) {
  if (ext) sprintf(dest, "extension");
  else dest[0] = '\0';
}


struct netExt {
  int removeMe;
};

/* This is how you register fields so they are visible in the object viewer 
   and accessible with setObject and getObject */
static void initNetExtInfo(void) {
  /*
  NetExt NE;
  addObj(NetExtInfo, "removeMe", OFFSET(NE, removeMe), TRUE, IntInfo);
   */
  NetExt NX;
  addObj(NetExtInfo, "RemoveMe", OFFSET(NX, removeMe), TRUE, IntInfo);
}

/* Use this to build the network extension and set your own default 
   parameters */
flag initNetworkExtension(Network N) {
  FREE(N->ext);

  N->ext = safeCalloc(1, sizeof(struct netExt), 
     "initNetworkExtension:N->ext");
  N->ext->removeMe = 3;

  return TCL_OK;
}

/* This is called when the network is reset.  If you also want to reset group
   or other extensions, you should do that from within here. */
flag resetNetworkExtension(Network N) {
  return TCL_OK;
}

flag freeNetworkExtension(Network N) {
  FREE(N->ext);
  return TCL_OK;
}


struct groupExt {
  int removeMe;
};

static void initGroupExtInfo(void) {
  /* GroupExt GE; */
}

flag initGroupExtension(Group G) {
  FREE(G->ext);
  return TCL_OK;
}

flag freeGroupExtension(Group G) {
  FREE(G->ext);
  return TCL_OK;
}


struct unitExt {
  int removeMe;
};

static void initUnitExtInfo(void) {
  /* UnitExt UE; */
}

flag initUnitExtension(Unit U) {
  FREE(U->ext);
  return TCL_OK;
}

flag freeUnitExtension(Unit U) {
  FREE(U->ext);
  return TCL_OK;
}


struct blockExt {
  int removeMe;
};

static void initBlockExtInfo(void) {
  /* BlockExt BE; */
}

flag initBlockExtension(Block B) {
  FREE(B->ext);
  return TCL_OK;
}

flag freeBlockExtension(Block B) {
  FREE(B->ext);
  return TCL_OK;
}

flag copyBlockExtension(Block from, Block to) {
  if (!from->ext) {
    if (to->ext) {
      freeBlockExtension(to);
      to->ext = NULL;
    }
    return TCL_OK;
  }
  if (!to->ext) initBlockExtension(to);
  to->ext->removeMe = from->ext->removeMe;
  return TCL_OK;
}


struct exSetExt {
  int removeMe;
};

static void initExSetExtInfo(void) {
  /* ExSetExt SE; */
}

flag initExSetExtension(ExampleSet S) {
  FREE(S->ext);
  return TCL_OK;
}

flag freeExSetExtension(ExampleSet S) {
  FREE(S->ext);
  return TCL_OK;
}


struct exampleExt {
  int removeMe;
};

static void initExampleExtInfo(void) {
  /* ExampleExt EE; */
}

flag initExampleExtension(Example E) {
  FREE(E->ext);
  return TCL_OK;
}

flag freeExampleExtension(Example E) {
  FREE(E->ext);
  return TCL_OK;
}


struct eventExt {
  int removeMe;
};

static void initEventExtInfo(void) {
  /* EventExt VE; */
}

flag initEventExtension(Event V) {
  FREE(V->ext);
  return TCL_OK;
}

flag freeEventExtension(Event V) {
  FREE(V->ext);
  return TCL_OK;
}


/********************************* New Tcl Commands **************************/
/* Here is a sample Tcl-callable C stub.  util.c contains the functions
   warning(), result(), and append() for creating Tcl return messages.
   The stub must be registered (see below) before it can be called. */


static int C_energy(TCL_CMDARGS) {
  char *usage = "energy <group A> <group B>";
  if (argc == 2 && !strcmp(argv[1], "-h"))
    return commandHelp(argv[0]);
  if (argc != 3) return usageError(argv[0], usage);
  if (!Net) return warning("%s: no current network", argv[0]);
  
  Group G1, G2;

  if( !(G1 = lookupGroup( argv[1] ) ) ) return warning( "could not find group %s", argv[1] );
  if( !(G2 = lookupGroup( argv[2] ) ) ) return warning( "could not find group %s", argv[2] );

  Unit U1, U2; Unit sU1, sU2;
  Link L;

  real energy = 0;

  for( U1 = G1->unit, sU1 = U1 + G1->numUnits; U1 < sU1; U1++ ) {
    for( U2 = G2->unit, sU2 = U2 + G2->numUnits; U2 < sU2; U2++ ) {
      if( (L = lookupLink( U1, U2, NULL, ALL_LINKS ) ) )
	energy += L->weight * U1->output * U2->output;
      if( (L = lookupLink( U2, U1, NULL, ALL_LINKS ) ) )
	energy += L->weight * U1->output * U2->output;
    }
  }

  return result( "%f", energy );
}

static int C_energy3(TCL_CMDARGS) {
  char *usage = "energy3 <group A> <group B> <group C>";
  if (argc == 2 && !strcmp(argv[1], "-h"))
    return commandHelp(argv[0]);
  if (argc != 4) return usageError(argv[0], usage);
  if (!Net) return warning("%s: no current network", argv[0]);
  
  Group G1, G2, G3;

  if( !(G1 = lookupGroup( argv[1] ) ) ) return warning( "could not find group %s", argv[1] );
  if( !(G2 = lookupGroup( argv[2] ) ) ) return warning( "could not find group %s", argv[2] );
  if( !(G3 = lookupGroup( argv[3] ) ) ) return warning( "could not find group %s", argv[3] );

  Unit U1, U2, U3; Unit sU1, sU2, sU3;
  Link L1, L2;

  real energy = 0, u1o, u2sq, u3o;
  char type01 = 0;
  if( G3->minOutput == 0 ) type01 = 1;

  for( U1 = G1->unit, sU1 = U1 + G1->numUnits; U1 < sU1; U1++ ) {
    if( U1->output != 0 ) {
      for( U2 = G2->unit, sU2 = U2 + G2->numUnits; U2 < sU2; U2++ ) {
	for( U3 = G3->unit, sU3 = U3 + G3->numUnits; U3 < sU3; U3++ ) {
	  L1 = lookupLink( U1, U2, NULL, ALL_LINKS );
	  L2 = lookupLink( U2, U3, NULL, ALL_LINKS );
	  if( L1 && L2 ) {
	    u1o = U1->output;
	    u2sq = U2->output * U2->output;
	    u3o = (type01 ? U3->output * 2 - 1 : U3->output);
	    energy += u1o * L1->weight * u2sq * L2->weight * u3o;
	  }
	}
      }
    }
  }

  return result( "%f", energy );
}

static int C_energy3b(TCL_CMDARGS) {
  char *usage = "energy3b <group A> <group B> <group C>";
  if (argc == 2 && !strcmp(argv[1], "-h"))
    return commandHelp(argv[0]);
  if (argc != 4) return usageError(argv[0], usage);
  if (!Net) return warning("%s: no current network", argv[0]);
  
  Group G1, G2, G3;

  if( !(G1 = lookupGroup( argv[1] ) ) ) return warning( "could not find group %s", argv[1] );
  if( !(G2 = lookupGroup( argv[2] ) ) ) return warning( "could not find group %s", argv[2] );
  if( !(G3 = lookupGroup( argv[3] ) ) ) return warning( "could not find group %s", argv[3] );

  Unit U1, U2, U3; Unit sU1, sU2, sU3;
  Link L1, L2;

  real energy = 0, u1o, u2sq, u3o;
  char type01 = 0; int u3 = 0;
  if( G3->minOutput == 0 ) type01 = 1;

  for( U1 = G1->unit, sU1 = U1 + G1->numUnits, U3 = G3->unit; U1 < sU1; U1++ ) {
    if( U1->output != 0 ) {
      for( U2 = G2->unit, sU2 = U2 + G2->numUnits; U2 < sU2; U2++ ) {
	  L1 = lookupLink( U1, U2, NULL, ALL_LINKS );
	  L2 = lookupLink( U2, U3, NULL, ALL_LINKS );
	  if( L1 && L2 ) {
	    u1o = U1->output;
	    u2sq = U2->output * U2->output;
	    u3o = (type01 ? U3->output * 2 - 1 : U3->output);
	    energy += u1o * L1->weight * u2sq * L2->weight * u3o;
	  }
      }
    }
    U3++; u3++; if( u3 >= G3->numUnits ) U3 = G3->unit;
  }

  return result( "%f", energy );
}


 
/******************************* User Initialization *************************/

/* This gets called while Lens is booting up */
flag userInit(void) {
  NetExtInfo = newObject("Network Extension", sizeof(struct netExt), 1,
			 extensionName, NULL, NULL);
  GroupExtInfo = newObject("Group Extension", sizeof(struct groupExt), 3,
			   extensionName, NULL, NULL);
  UnitExtInfo = newObject("Unit Extension", sizeof(struct unitExt), 5,
			  extensionName, NULL, NULL);
  BlockExtInfo = newObject("Block Extension", sizeof(struct blockExt), 7,
			   extensionName, NULL, NULL);
  ExSetExtInfo = newObject("Example Set Extension", sizeof(struct exSetExt), 3,
			   extensionName, NULL, NULL);
  ExampleExtInfo = newObject("Example Extension", sizeof(struct exampleExt), 5,
			     extensionName, NULL, NULL);
  EventExtInfo = newObject("Event Extension", sizeof(struct eventExt), 7,
			   extensionName, NULL, NULL);
  initNetExtInfo();
  initGroupExtInfo();
  initUnitExtInfo();
  initBlockExtInfo();
  initExSetExtInfo();
  initExampleExtInfo();
  initEventExtInfo();

  /* This creates a blank line in the list of commands produced by "help": */
  registerCommand(NULL, "", ""); 
  /* This creates the command "foo": */
  registerCommand(C_energy, "energy", "calculate energy-like measure between two groups"); 
  registerCommand(C_energy3, "energy3", "calculate energy-like measure between three groups"); 
  registerCommand(C_energy3b, "energy3b", "calculate energy-like measure between three groups"); 

  registerGroupType("TANH_ENTROPY", TANH_ENTROPY, GROUP_COST,
		    tanhEntropyErrorInit);

  return TCL_OK;
}

void tanhEntropyError(Group G, GroupProc P) {
  real output, target, error = 0.0, 
    targetRadius = chooseValue(G->targetRadius, Net->targetRadius),
    zeroErrorRadius = chooseValue(G->zeroErrorRadius, Net->zeroErrorRadius);
  if (targetRadius != 0.0 || zeroErrorRadius != 0.0) {
    FOR_EACH_UNIT(G, {
      if (!isNaN(U->target)) {
	output = (U->output + 1.0) / 2.0;  
	target = (U->target + 1.0) / 2.0;
	target = U->adjustedTarget = 
	  ADJUSTED_TARGET(output, target, targetRadius, zeroErrorRadius);
	error += CROSS_ENTROPY_ERROR(output, target);
		  
      }
    });
  } else {
    FOR_EACH_UNIT(G, {
      if (!isNaN(U->target)) {
	output = (U->output + 1.0) / 2.0;
	target = U->adjustedTarget = (U->target + 1.0) / 2.0;
	error += CROSS_ENTROPY_ERROR(output, target);
		  
      }
    });
  }
  error *= Net->dt * G->dtScale * G->errorScale * 
    ((Net->pseudoExampleFreq) ? Net->currentExample->frequency : 1.0);
  G->error += error;
  Net->error += error;
}

void tanhEntropyErrorBack(Group G, GroupProc P) {
  real scale = ((Net->pseudoExampleFreq) ? Net->currentExample->frequency :
		1.0) * G->errorScale, output, target;
  FOR_EACH_UNIT(G, {
    if (!isNaN(U->target)) {
      output = (U->output + 1.0) / 2.0; 
      target = U->adjustedTarget;
      U->outputDeriv += scale * CROSS_ENTROPY_DERIV(output, target);
	
    }
  });
}

void tanhEntropyErrorInit(Group G, GroupProc P) {
  P->forwardProc  = tanhEntropyError;
  P->backwardProc = tanhEntropyErrorBack;
}
