#include "system.h"
#include "util.h"
#include "type.h"
#include "network.h"
#include "connect.h"
#include "example.h"
#include "train.h"
#include "act.h"
#include "command.h"
#include "control.h"
#include "object.h"

/* Add your initials or something here to customize the version number */
char *VersionExt = "CTK";

/******************************** Extension Records **************************/

ObjInfo NetExtInfo;
ObjInfo GroupExtInfo;
ObjInfo UnitExtInfo;
ObjInfo BlockExtInfo;
ObjInfo ExSetExtInfo;
ObjInfo ExampleExtInfo;
ObjInfo EventExtInfo;

/* This is what the object viewer displays as the name of the extension. */
static void extensionName(void *ext, char *dest) {
  if (ext) sprintf(dest, "extension");
  else dest[0] = '\0';
}


struct netExt {
  int removeMe;
};

/* This is how you register fields so they are visible in the object viewer 
   and accessible with setObject and getObject */
static void initNetExtInfo(void) {
  /*
  NetExt NE;
  addObj(NetExtInfo, "removeMe", OFFSET(NE, removeMe), TRUE, IntInfo);
   */
  NetExt NX;
  addObj(NetExtInfo, "RemoveMe", OFFSET(NX, removeMe), TRUE, IntInfo);
}

/* Use this to build the network extension and set your own default 
   parameters */
flag initNetworkExtension(Network N) {
  FREE(N->ext);

  N->ext = safeCalloc(1, sizeof(struct netExt), 
     "initNetworkExtension:N->ext");
  N->ext->removeMe = 3;

  return TCL_OK;
}

/* This is called when the network is reset.  If you also want to reset group
   or other extensions, you should do that from within here. */
flag resetNetworkExtension(Network N) {
  return TCL_OK;
}

flag freeNetworkExtension(Network N) {
  FREE(N->ext);
  return TCL_OK;
}


struct groupExt {
  int removeMe;
};

static void initGroupExtInfo(void) {
  /* GroupExt GE; */
}

flag initGroupExtension(Group G) {
  FREE(G->ext);
  return TCL_OK;
}

flag freeGroupExtension(Group G) {
  FREE(G->ext);
  return TCL_OK;
}


struct unitExt {
  int removeMe;
};

static void initUnitExtInfo(void) {
  /* UnitExt UE; */
}

flag initUnitExtension(Unit U) {
  FREE(U->ext);
  return TCL_OK;
}

flag freeUnitExtension(Unit U) {
  FREE(U->ext);
  return TCL_OK;
}


struct blockExt {
  int removeMe;
};

static void initBlockExtInfo(void) {
  /* BlockExt BE; */
}

flag initBlockExtension(Block B) {
  FREE(B->ext);
  return TCL_OK;
}

flag freeBlockExtension(Block B) {
  FREE(B->ext);
  return TCL_OK;
}

flag copyBlockExtension(Block from, Block to) {
  if (!from->ext) {
    if (to->ext) {
      freeBlockExtension(to);
      to->ext = NULL;
    }
    return TCL_OK;
  }
  if (!to->ext) initBlockExtension(to);
  to->ext->removeMe = from->ext->removeMe;
  return TCL_OK;
}


struct exSetExt {
  int removeMe;
};

static void initExSetExtInfo(void) {
  /* ExSetExt SE; */
}

flag initExSetExtension(ExampleSet S) {
  FREE(S->ext);
  return TCL_OK;
}

flag freeExSetExtension(ExampleSet S) {
  FREE(S->ext);
  return TCL_OK;
}


struct exampleExt {
  int removeMe;
};

static void initExampleExtInfo(void) {
  /* ExampleExt EE; */
}

flag initExampleExtension(Example E) {
  FREE(E->ext);
  return TCL_OK;
}

flag freeExampleExtension(Example E) {
  FREE(E->ext);
  return TCL_OK;
}


struct eventExt {
  int removeMe;
};

static void initEventExtInfo(void) {
  /* EventExt VE; */
}

flag initEventExtension(Event V) {
  FREE(V->ext);
  return TCL_OK;
}

flag freeEventExtension(Event V) {
  FREE(V->ext);
  return TCL_OK;
}


/********************************* New Tcl Commands **************************/
/* Here is a sample Tcl-callable C stub.  util.c contains the functions
   warning(), result(), and append() for creating Tcl return messages.
   The stub must be registered (see below) before it can be called. */
/*
static int C_foo(TCL_CMDARGS) {
  char *usage = "foo <bar>";
  if (argc == 2 && !strcmp(argv[1], "-h"))
    return commandHelp(argv[0]);
  if (argc != 2) return usageError(argv[0], usage);
  if (!Net) return warning("%s: no current network", argv[0]);
  
  return TCL_OK;
}
*/

 
/******************************* User Initialization *************************/

/* This gets called while Lens is booting up */
flag userInit(void) {
  NetExtInfo = newObject("Network Extension", sizeof(struct netExt), 1,
			 extensionName, NULL, NULL);
  GroupExtInfo = newObject("Group Extension", sizeof(struct groupExt), 3,
			   extensionName, NULL, NULL);
  UnitExtInfo = newObject("Unit Extension", sizeof(struct unitExt), 5,
			  extensionName, NULL, NULL);
  BlockExtInfo = newObject("Block Extension", sizeof(struct blockExt), 7,
			   extensionName, NULL, NULL);
  ExSetExtInfo = newObject("Example Set Extension", sizeof(struct exSetExt), 3,
			   extensionName, NULL, NULL);
  ExampleExtInfo = newObject("Example Extension", sizeof(struct exampleExt), 5,
			     extensionName, NULL, NULL);
  EventExtInfo = newObject("Event Extension", sizeof(struct eventExt), 7,
			   extensionName, NULL, NULL);
  initNetExtInfo();
  initGroupExtInfo();
  initUnitExtInfo();
  initBlockExtInfo();
  initExSetExtInfo();
  initExampleExtInfo();
  initEventExtInfo();

  /* This creates a blank line in the list of commands produced by "help": */
  /* registerCommand(NULL, "", ""); */
  /* This creates the command "foo": */
  /* registerCommand(C_foo, "foo", "this command does some stuff"); */

  registerGroupType("TANH_ENTROPY", TANH_ENTROPY, GROUP_COST,
		    tanhEntropyErrorInit);

  return TCL_OK;
}

void tanhEntropyError(Group G, GroupProc P) {
  real output, target, error = 0.0, 
    targetRadius = chooseValue(G->targetRadius, Net->targetRadius),
    zeroErrorRadius = chooseValue(G->zeroErrorRadius, Net->zeroErrorRadius);
  if (targetRadius != 0.0 || zeroErrorRadius != 0.0) {
    FOR_EACH_UNIT(G, {
      if (!isNaN(U->target)) {
	output = (U->output + 1.0) / 2.0;  
	target = (U->target + 1.0) / 2.0;
	target = U->adjustedTarget = 
	  ADJUSTED_TARGET(output, target, targetRadius, zeroErrorRadius);
	error += CROSS_ENTROPY_ERROR(output, target);
		  
      }
    });
  } else {
    FOR_EACH_UNIT(G, {
      if (!isNaN(U->target)) {
	output = (U->output + 1.0) / 2.0;
	target = U->adjustedTarget = (U->target + 1.0) / 2.0;
	error += CROSS_ENTROPY_ERROR(output, target);
		  
      }
    });
  }
  error *= Net->dt * G->dtScale * G->errorScale * 
    ((Net->pseudoExampleFreq) ? Net->currentExample->frequency : 1.0);
  G->error += error;
  Net->error += error;
}

void tanhEntropyErrorBack(Group G, GroupProc P) {
  real scale = ((Net->pseudoExampleFreq) ? Net->currentExample->frequency :
		1.0) * G->errorScale, output, target;
  FOR_EACH_UNIT(G, {
    if (!isNaN(U->target)) {
      output = (U->output + 1.0) / 2.0; 
      target = U->adjustedTarget;
      U->outputDeriv += scale * CROSS_ENTROPY_DERIV(output, target);
	
    }
  });
}

void tanhEntropyErrorInit(Group G, GroupProc P) {
  P->forwardProc  = tanhEntropyError;
  P->backwardProc = tanhEntropyErrorBack;
}

flag autoseqNetRunExample(Example E, flag (*tickProc)(Event), flag *correct) {
  int tick = 0, ticksOnEvent, event = 0;
  flag eventDone, critReached, training = (tickProc == Net->netTrainTick);
  real timeOnEvent, minTime, maxTime, graceTime;
  Event V;

  if (E->set->loadEvent(V = E->event)) return TCL_ERROR;
  ticksOnEvent = 0;
  maxTime = chooseValue(V->maxTime, E->set->maxTime);
  minTime = chooseValue(V->minTime, E->set->minTime);
  graceTime = chooseValue(V->graceTime, E->set->graceTime);
  Net->inGracePeriod = (graceTime > 0.0);
  
  /* Reset the outputs and integrators if requested. */
  FOR_EACH_GROUP(if (G->type & RESET_ON_EXAMPLE) {
    resetOutputs(G);
    resetForwardIntegrators(G);
  });
  
  Net->exampleHistoryStart = 
    (Net->exampleHistoryStart + Net->ticksOnExample) % Net->historyLength;

  Net->resetHistory[HISTORY_INDEX(0)] = TRUE;

  RUN_PROC(preEventProc);
  *correct = TRUE;
  for (; tick < Net->maxTicks && event < E->numEvents / 2; tick++) {
    Net->currentTick = tick;
    Net->eventHistory[tick] = event;
    if (tick > 0) Net->resetHistory[HISTORY_INDEX(tick)] = FALSE;
    
    RUN_PROC(preTickProc);

    if (srbttNetForward(V)) return TCL_ERROR;   /* This is the important call */

    RUN_PROC(postTickProc);
    
    updateDisplays(ON_TICK);
    
    eventDone = FALSE;
    critReached = 2;
    ticksOnEvent++;
    /* The SMALL_VAL is to correct for possible floating point inaccuracy */
    timeOnEvent = ((real) ticksOnEvent / Net->ticksPerInterval) + SMALL_VAL;
    /* If we have done enough time on this event, go on */
    if (timeOnEvent >= maxTime || tick >= Net->maxTicks - 1)
      eventDone = TRUE;
    /* If we have gone long enough and reached criterion, go on */
    else if (timeOnEvent >= minTime && 
	     (critReached = groupCriteriaReached(training)))
      eventDone = TRUE;
    /* See if the grace period is over */
    else if (Net->inGracePeriod && timeOnEvent >= graceTime)
      Net->inGracePeriod = FALSE;

    if (eventDone) {
      if (critReached == 2) critReached = groupCriteriaReached(training);
      if (!critReached) *correct = FALSE;
      RUN_PROC(postEventProc);
      updateDisplays(ON_EVENT);

      if (Net->groupCritRequired && !critReached) break;

      if (++event < E->numEvents && ((V = E->event + event))) {
	if (E->set->loadEvent(V)) return TCL_ERROR;
	ticksOnEvent = 0;
	maxTime = chooseValue(V->maxTime, E->set->maxTime);
	minTime = chooseValue(V->minTime, E->set->minTime);
	graceTime = chooseValue(V->graceTime, E->set->graceTime);
	Net->inGracePeriod = (graceTime > 0.0);
	RUN_PROC(preEventProc);
      }
    }
  }
  Net->ticksOnExample = Net->currentTick + 1;

  updateDisplays(ON_EXAMPLE);
  return TCL_OK;
}

flag autoseqNetRunExample(Example E, flag (*tickProc)(Event), flag *correct) {
  int tick = 0, ticksOnEvent, event = 0;
  flag eventDone, critReached, training = (tickProc == Net->netTrainTick);
  real timeOnEvent, minTime, maxTime, graceTime;
  Event V;

  if (E->set->loadEvent(V = E->event)) return TCL_ERROR;
  ticksOnEvent = 0;
  maxTime = chooseValue(V->maxTime, E->set->maxTime);
  minTime = chooseValue(V->minTime, E->set->minTime);
  graceTime = chooseValue(V->graceTime, E->set->graceTime);
  Net->inGracePeriod = (graceTime > 0.0);
  
  /* Reset the outputs and integrators if requested. */
  FOR_EACH_GROUP(if (G->type & RESET_ON_EXAMPLE) {
    resetOutputs(G);
    resetForwardIntegrators(G);
  });
  
  if (Net->type & CONTINUOUS) {
    storeOutputsAndTargets(0);
    Net->exampleHistoryStart = 0;
    Net->eventHistory[tick] = -1;
    tick = 1;
    updateDisplays(ON_TICK);
  } else {
    Net->exampleHistoryStart = 
      (Net->exampleHistoryStart + Net->ticksOnExample) % Net->historyLength;
  }
  Net->resetHistory[HISTORY_INDEX(0)] = TRUE;

  RUN_PROC(preEventProc);
  *correct = TRUE;
  for (; tick < Net->maxTicks && event < E->numEvents; tick++) {
    Net->currentTick = tick;
    Net->eventHistory[tick] = event;
    if (tick > 0) Net->resetHistory[HISTORY_INDEX(tick)] = FALSE;
    
    RUN_PROC(preTickProc);
    if (tickProc(V)) return TCL_ERROR;   /* This is the important call */
    RUN_PROC(postTickProc);
    
    updateDisplays(ON_TICK);
    
    eventDone = FALSE;
    critReached = 2;
    ticksOnEvent++;
    /* The SMALL_VAL is to correct for possible floating point inaccuracy */
    timeOnEvent = ((real) ticksOnEvent / Net->ticksPerInterval) + SMALL_VAL;
    /* If we have done enough time on this event, go on */
    if (timeOnEvent >= maxTime || tick >= Net->maxTicks - 1)
      eventDone = TRUE;
    /* If we have gone long enough and reached criterion, go on */
    else if (timeOnEvent >= minTime && 
	     (critReached = groupCriteriaReached(training)))
      eventDone = TRUE;
    /* See if the grace period is over */
    else if (Net->inGracePeriod && timeOnEvent >= graceTime)
      Net->inGracePeriod = FALSE;

    if (eventDone) {
      if (critReached == 2) critReached = groupCriteriaReached(training);
      if (!critReached) *correct = FALSE;
      RUN_PROC(postEventProc);
      updateDisplays(ON_EVENT);

      if (Net->groupCritRequired && !critReached) break;

      if (++event < E->numEvents && ((V = E->event + event))) {
	if (E->set->loadEvent(V)) return TCL_ERROR;
	ticksOnEvent = 0;
	maxTime = chooseValue(V->maxTime, E->set->maxTime);
	minTime = chooseValue(V->minTime, E->set->minTime);
	graceTime = chooseValue(V->graceTime, E->set->graceTime);
	Net->inGracePeriod = (graceTime > 0.0);
	RUN_PROC(preEventProc);
      }
    }
  }
  Net->ticksOnExample = Net->currentTick + 1;

  updateDisplays(ON_EXAMPLE);
  return TCL_OK;
}


flag CTKsrbpttNetForward(Event V) {
  FOR_EACH_GROUP({
    computeInput(G, TRUE);
    computeOutput(G, TRUE);
    computeCost(G, TRUE);
    computeCostBack(G, TRUE);
  });
  return TCL_OK;
}

flag CTKsrbpttNetExampleBack(Example E) {
  /* Clear the output deriv caches. */
  /* If integrating, reset the deriv accumulators. */
  FOR_EACH_GROUP(resetOutputDerivCache(G); resetBackwardIntegrators(G));

  FOR_EACH_GROUP_BACK(BACKPROP(G));

  for (Net->currentTick = Net->ticksOnExample - 2; 
       Net->currentTick >= 0; Net->currentTick--) {
    /* Set outputDerivs to the stored instant error derivatives. */
    FOR_EACH_GROUP({
      restoreOutputDerivs(G, Net->currentTick);
      restoreOutputs(G, Net->currentTick);
      restoreInputs(G, Net->currentTick);
    });
    FOR_EACH_GROUP_BACK(BACKPROP(G));
  }
  /* Restore the final outputs for continuity with the next example */
  FOR_EACH_GROUP(restoreOutputs(G, Net->ticksOnExample - 1));
  return TCL_OK;
}

